(* ::Package:: *)

BeginPackage["BenderWu`"];


(* ::InheritFromParent:: *)
(*"BenderWu`"*)


BWDifference::usage = "The BWDifference computes the perturbative expansion of the difference equation."
BenderWu::usage = "BenderWu function version 1.0. For bug report and suggestions, please contact Tin Sulejmanpasic at tin.sulejmanpasic@gmail.com

The BenderWu function calculates the solution of the differential equation

-(1/2)\[Psi]''(x)+1/g^2 \*SubscriptBox[V, 0](gx)\[Psi](x)+\*SubscriptBox[V, 2](gx)\[Psi](x)=\[Epsilon]\[Psi](x)

as a parturbation series in the coupling g, i.e.

\[Psi](x)=\!\(
\*UnderoverscriptBox[\(\[Sum]\), \(l = 0\), \(\[Infinity]\)]\(
\*SuperscriptBox[
SubscriptBox[\(A\), \(l\)], \(k\)] 
\*SuperscriptBox[\(x\), \(k\)] 
\*SuperscriptBox[\(g\), \(l\)]\)\) ,       \[Epsilon] = \!\(
\*UnderoverscriptBox[\(\[Sum]\), \(n = 0\), \(\[Infinity]\)]
\*SubscriptBox[\(\[Epsilon]\), \(2  n\)]\)

Notice that the expansion of \[Epsilon] is only in powers of g^2. 

The call to the function BenderWu has the following syntax

BenderWu[{\*SubscriptBox[V, \(0\)][x],\(\*SubscriptBox[V, \(2\)][x]},x,\[Nu],\*SubscriptBox[n, \(max\)],Options]

or 

BenderWu[Subscript[V, 0][x],x,\[Nu],\*SubscriptBox[n,max],Options] (for \*SubscriptBox[V, 2][x]=0)

The respectivelly specify the potentials \*SubscriptBox[V, \(0\)][x] and \*SubscriptBox[V, \(2\)][x], the argument of th epotential (x in this case), the harmonic oscillator level number \[Nu], the maximal order of g^2 and a number of options (see below)

If called without any visualisation options, the function BenderWu returns a list of three elements. The first element gives a list of the perturbative corrections \*SubscriptBox[\[Epsilon], \(2  n\)]

{\*SubscriptBox[\[Epsilon], 0],\*SubscriptBox[\[Epsilon], 2],\*SubscriptBox[\[Epsilon], 4],..., \*SubscriptBox[\[Epsilon], \(2\*SubscriptBox[n,max]\)]}

The second element returns a matrix

\*SuperscriptBox[SubscriptBox[A, \(l\)], \(k\)],      l=0,1,...,2*n_max, k=0,1,...,\[Nu]+3*(2*n_max)

while the third element returns a list of four elements required for processing by the function BWProcess.

Options and their possible values are (default value listed first): 
Evaluation->\"Analytical\", Evaluation->\"Numerical\",
WorkingPrecision->MachinePrecision (The working precision if the Evaluation is \"Numerical\")
Coupling->g (A symbol to be used for coupling in output)
Monitor->True,Monitor->False (Enable or disable the computation monitor) 
Simplify->False,Simplify->True (Simplify the expressions at intermediate steps - Turning this on may speed up analytic evaluations)
MaxTime->Infinity, (time in seconds allowed for a single order computation.)
TotalMaxTime->Infinity, (time in minutes allowed to generate the series.)

The BenderWu function calles the function BWProcess to generate the output, and therefore it also takes in options shared by the function BWProcess. These are called the visaulisation options.

Output->\"BWProcess\", Output->\"Energy\" (output energy), Output->\"WaveFunction\" (output wave-function)
OutputStyle->\"Array\", OutputStyle->\"Series\" (display the output as a series in the coupling), OutputStyle->\"MatrixForm\" (display output as a matrix),
SeriesOutput->\"Coupling\" (group coupling in wave-function) ,SeriesOutput->\"Argument\" (group powers of argument x in wave-function)";
BWProcess::usage = "A process function for the BenderWu package. It takes in a raw output from the BenderWu function without any visualisation options, and formats the raw data from the BenderWu function for output. It takes";
BWLevelPolynomial::usage = "Takes a two matrix \!\(\*SubscriptBox[\(\[Epsilon]\), \(l\)]\)(\[Nu]) of perturbative energy corrections at order l, and for energy level \[Nu] and returns a list of polynomials in \[Nu] for l."
LevelPoly::usage = "A function which takes in a list of numbers, \!\(\*SubscriptBox[\(a\), \(\[Nu]\)]\), and fits a polynomial \!\(\*SubscriptBox[\(c\), \(0\)]\)+\!\(\*SubscriptBox[\(c\), \(1\)]\)\[Nu]+\!\(\*SubscriptBox[\(c\), \(2\)]\)\!\(\*SuperscriptBox[\(\[Nu]\), \(2\)]\)+..." 


\[Nu]; (*global variables*)
Begin["`Private`"]
SetAttributes[BenderWu,HoldAll];
optmain={OutputStyle->"Array",SeriesOutput->"Coupling",Evaluation->"Analytical",Coupling->"g",
HighestPower->0,ArrayForm->False,Echo->False,Monitor->True};
Options[BenderWu]=Join[{Output->"BWProcess",WorkingPrecision->MachinePrecision,
MaxTime->Infinity,TotalMaxTime->Infinity,Warnings->True,Simplify->False},optmain];

SyntaxInformation[BenderWu]={"ArgumentsPattern"->{_,_,_,_,OptionsPattern[]},"LocalVariables"->{"Solve",{2}}};

BenderWu[Potential_,x_,nu_,ord_,OptionsPattern[]]:=
Module[
(*Local variables:*)
{l,vf,v,A,epsilon,omega,L,lstep,evencheck,V,Vf,TimeTemp,even,MonitorDelay,k,e0,i,coupling,lmax,vmax,vfmax,timesteptemp},

(*Echo message start*)
If[OptionValue[Echo]==True, (*Echo messages*)
	Print["BenderWu function ver 1.0"];
	Print["To turn off these messages use the option Echo->False. \nTo turn off monitoring use option Monitor->False."];
];
(*Check if level number is an integer*)
If[TrueQ[Head[nu]=!=Integer],Print["The level number must be an integer. Aborting..."];Return[];];

(*Sets the maximum value of l, counting order of g*)
lmax=2(ord); 
(*If the coupling is given as a string, convert it to an expression*)
If[TrueQ[Head[OptionValue[Coupling]]===String],
	coupling=ToExpression[OptionValue[Coupling]],
	coupling=OptionValue[Coupling]
	];
If[TrueQ[Head[Potential]==List],
		If[Length[Potential]==2,
	(*then*)		V=Potential[[1]];
					Vf=Potential[[2]];,
	(*else*)		Print["Error in the input of the potential! Input: BenderWu[V[x],x... or BenderWu[{V0[x],V2[x]},x]"]; Return[$Aborted];
			];,
	V=Potential;
	Vf=0;
];
omega=Simplify[Sqrt[D[V,{x,2}]/.x->0]];
If[omega==0, Print["The quadratic term of the potential is zero. Aborting..."]; Return[$Aborted];];
e0=omega(nu+1/2)+Vf/.x->0; (*energy to leading order*)
v=Table[Evaluate[D[V/omega^2,{x,(i+2)}]/((i+2)!)]/.x->0,{i,1,lmax}]; (*coefficients of V[x]*)
vf=Table[Evaluate[D[Vf/omega,{x,i}]]/(i!)/.x->0,{i,1,lmax}]; (*Coefficients of Vf[x]*)

If[TrueQ[v[[Length[v]]]=!=0],
vmax=Length[v];,
vmax=Length[Flatten[Drop[Split[v],-1]]];
];
If[TrueQ[vf[[Length[vf]]]=!=0],
vfmax=Length[vf];,
vfmax=Length[Flatten[Drop[Split[vf],-1]]];
];
even=0;
evencheck=Sum[v[[i]]^2+vf[[i]]^2,{i,1,lmax,2}];
If[TrueQ[evencheck==0],
		lstep=2; 
		even=1;,
		lstep=1; 
		even=0;
	];
A=Table[0,{l,0,lmax},{k,0,nu+3lmax+2}]; 
epsilon=Table[0,{i,1,Floor[lmax/2]}]; (*Sub-leading energy*)
If[OptionValue[Evaluation]=="Numerical",A[[1,nu+1]]=N[1,OptionValue[WorkingPrecision]];,A[[1,nu+1]]=1;]; (*Normalization*)
timesteptemp=1;
timeratio=0;
(*Monitor[*)
BeginingOfTime=AbsoluteTime[];
CPUtimestep=0;
k=0;
If[(OptionValue[Monitor]==True)&&NumericQ[(l*k*CPUtimestep*TimeLapsed)]==True,
	MonitorCell=PrintTemporary["Computing order: ",Dynamic[Floor[l/2]-1],"\nComputing polynomial power: ",Dynamic[k],"\nCPU time per order: ",Dynamic[Floor[2 CPUtimestep 10]/10.0]," seconds.\nMemory in use: ",Dynamic[Floor[MemoryInUse[]/(1024^2)]]," MB","\nTotal time elapsed: ",Dynamic[Floor[(TimeLapsed)/60]]," minutes and ",Dynamic[Floor[Mod[(TimeLapsed),60]]]," seconds."];
];
tmtot=TimeConstrained[
(*Monitor[*)
	For[l=0,l<=lmax,l+=lstep,
		ClearSystemCache[];
		If[l==240,SetSystemOptions["CompileOptions" -> {"SumCompileLength" -> lmax+10}]];
		tm=TimeConstrained[
		CPUtimebeg=TimeUsed[];
		timebeg=AbsoluteTime[];
		TimeLapsed=AbsoluteTime[]-BeginingOfTime;

		(*If[OptionValue[Evaluation]=="Numerical",A[[l+1,nu+1]]=N[1,OptionValue[WorkingPrecision]];,A[[l+1,nu+1]]=1;]*);
		If[l>0,
			For[k=nu+3l,k>nu,k--,
				A[[l+1,k+1]]=1/(2(k-nu))((k+2)(k+1)A[[l+1,k+2+1]]+Sum[2 epsilon[[n]]A[[l-2n+1,k+1]],{n,1,Floor[l/2]}]-2Sum[v[[n]]A[[l-n+1,k-n-2+1]],{n,1,Min[k-2,l,vmax]}]-2Sum[vf[[n]]A[[l-n+1,k-n+1]],{n,1,Min[k,l,vfmax]}]);
				If[OptionValue[Simplify]==True,A[[l+1,k+1]]=Simplify[A[[l+1,k+1]]]];	
				If[Mod[l,2]==0,
				epsilon[[l/2]]=-(nu+2)(nu+1)A[[l+1,nu+2+1]]/2+ Sum[v[[n]]A[[l-n+1,nu-n-2+1]],{n,1,Min[l,nu-2]}]+Sum[vf[[n]]A[[l-n+1,nu-n+1]],{n,1,Min[l,nu]}];
				If[OptionValue[Simplify]==True,epsilon[[l/2]]=Simplify[epsilon[[l/2]]]];
				];
			];
		];
		For[k=nu-1,k>=0,k--,
			A[[l+1,k+1]]=1/(2(k-nu))((k+2)(k+1)A[[l+1,k+2+1]]+Sum[2 epsilon[[n]]A[[l-2n+1,k+1]],{n,1,Floor[l/2]}]-2Sum[v[[n]]A[[l-n+1,k-n-2+1]],{n,1,Min[k-2,l]}]-2Sum[vf[[n]]A[[l-n+1,k-n+1]],{n,1,Min[k,l]}]);
			If[OptionValue[Simplify]==True,A[[l+1,k+1]]=Simplify[A[[l+1,k+1]]]];
			];

		CPUtimestep=(TimeUsed[]-CPUtimebeg);
		timestep=AbsoluteTime[]-timebeg;
		,OptionValue[MaxTime]];
		If[tm==$Aborted,
			lmax=2 Floor[(l-1)/2];
			If[lmax<1,Return[Print["Error... No orders were computed. Check that time constraint MaxTime is not too small."];];];
			If[OptionValue[Warnings]==True,
				Print["WARNING: Time constraint for a single order exeeded ",OptionValue[MaxTime], " seconds  at order ",coupling^l,"."];
				Print["Output generated to order ",Evaluate[coupling^(lmax)]];
			];
			Break[];
		];
	];
(*"Computing order: "<>ToString[Floor[l/2]-1]<>"\nComputing polynomial power: "<>ToString[k]<>"\nCPU time per order: "<>ToString[CPUtimestep]<>"\nMemory in use: "<>ToString[N[MemoryInUse[]/(1024^2)]]<>" MB"<>"\nTotal time elapsed: "<>ToString[Floor[(AbsoluteTime[]-BeginingOfTime)/60]]<>" minutes and "<>ToString[Floor[Mod[(AbsoluteTime[]-BeginingOfTime),60]]]<>" seconds.",
	MonitorDelay
];*)
,Evaluate[OptionValue[TotalMaxTime]60]];
	If[tmtot==$Aborted,
			lmax=2 Floor[(l-1)/2];
			If[lmax<1,Return[Print["Error... No orders were computed. Check that time constraint TotalMaxTime is not too small."];];];
			If[OptionValue[Warnings]==True,
			Print["WARNING: Time set to complete the evaluation ",N[OptionValue[TotalMaxTime],1], " minutes  was exceeded at order ",coupling^l,"."] 
			Print["Output generated to order ",Evaluate[coupling^(lmax)]];
			];
		];
NotebookDelete[MonitorCell];


If[OptionValue[Echo]==True,
	Print["Longest CPU time spend computing a single order is ", Floor[CPUtimestep 10]/10.0," seconds"];
	Print["Longest absolute time spend computing a single order is ",Floor[timestep 10]/10.0," seconds"];
	Print["Total time needed for the computation ",Floor[(AbsoluteTime[]-BeginingOfTime)/60]," minutes and ",Floor[Mod[(AbsoluteTime[]-BeginingOfTime),60] 10]/10.0," seconds"];
];
epsilon=Table[epsilon[[i]]/omega^i,{i,1,lmax/2}];
A=Table[A[[lp,k]]Sqrt[omega]^(k-lp),{lp,1,lmax+1},{k,1,nu+3lmax+1}]; 
(*Return[timestep];*)
Return[BWProcess[{Join[{Evaluate[V/.x->0]},{e0},omega epsilon],A,{x,nu,lmax,even}},Output->OptionValue[Output],OutputStyle->OptionValue[OutputStyle],SeriesOutput->OptionValue[SeriesOutput],Evaluation->OptionValue[Evaluation],Coupling->OptionValue[Coupling],HighestPower->OptionValue[HighestPower],ArrayForm->OptionValue[ArrayForm]]];
];


(*Begin Function BWProcess*)
Options[BWProcess]=Join[{Output->"Energy",Order->"Default",Warnings->True,Prefactor->False},optmain];
SetAttributes[BWProcess,HoldAll];
SyntaxInformation[BWProcess]={"ArgumentsPattern"->{_,OptionsPattern[]}};

BWProcess[All_,OptionsPattern[]]:=Module[{A,epsilon,x,nu,lmin,lmax,L,even,out,M,kmax,k,coupling,ratio,epstrunc,xi,alpha},
(*Note: even variable has nothing to do with even functions, and is used only to distinguish the BWDifference output and BenderWu output at this point*)
	If[TrueQ[Head[OptionValue[Coupling]]===String],coupling=ToExpression[OptionValue[Coupling]],coupling=OptionValue[Coupling]];
	
	If[TrueQ[Head[OptionValue[Order]]=!=List],
	(*then*)	If[TrueQ[OptionValue[Order]=="Default"],lmax=All[[3,3]];lmin=-2Boole[even==2],lmax=2OptionValue[Order]; lmin=-2Boole[even==2];]; (*Set the maximal order*),
	(*else*)    If[Length[OptionValue[Order]]!=2,Print["The order should be a non-negative integer or list of two non-negative integers. Aborting..."; Return[$Aborted]]];
				lmax=2 OptionValue[Order][[2]];
				lmin=2 OptionValue[Order][[1]];
	];

	If[lmin<-2,lmin=-2];
	If[lmax>All[[3,3]],lmax=All[[3,3]]];
	epsilon=Take[All[[1]],lmax/2+2];
	A=All[[2]];
	x=All[[3,1]];
	nu=All[[3,2]];
	even=All[[3,4]];
	If[Length[All[[3]]]>= 6,xi=All[[3,5]];alpha=All[[3,6]]];
(*	If[even\[Equal]2,lmin-=2;lmax-=0;];*)
	A=Take[A,lmax+1];
	A=Transpose[Take[Transpose[A],nu+3lmax+1]];
	If[OptionValue[Output]=="BWProcess", Return[All]];
	If[OptionValue[Output]=="Energy",
		If[OptionValue[OutputStyle]=="MatrixForm",Return[MatrixForm[Join[{Table[coupling^(2l),{l,lmin/2,(lmax)/2}]},{Drop[epsilon,lmin/2+1]}]]]];
		If[OptionValue[OutputStyle]=="Series",Return[Sum[epsilon[[i/2+2]]coupling^(i),{i,lmin,lmax,2}]];];
		If[OptionValue[OutputStyle]=="Ratio"||OptionValue[OutputStyle]=="RatioPlot",
		epstrunc=Drop[epsilon,lmin/2+1];
		If[Times@@epstrunc==0&&OptionValue[Warnings]==True,Print["WARNING: The energy expansion contains terms which are zero. These are dropped before constructing ratios."];];
		epstrunc=DeleteCases[epstrunc,0];
		ratio=Drop[epstrunc,1]/Drop[epstrunc,-1];
			];
		If[OptionValue[OutputStyle]=="RatioPlot",Return[ListPlot[ratio]];];
		If[OptionValue[OutputStyle]=="Ratio",Return[ratio];];
		Return[Drop[epsilon,lmin/2+1]];
	];
	If[OptionValue[Output]=="WaveFunction",
		If[lmin<0,lmin=0];
		If[OptionValue[OutputStyle]=="MatrixForm",
			kmax=Length[A[[1]]];
			M=Join[{Table[x^(i),{i,0,Length[A[[1]]]-1}]},A];
			M=Join[Transpose[{Join[{" "},Table[coupling^(i),{i,lmin,lmax}]]}],M,2];
			Return[MatrixForm[M]];
		];

		If[OptionValue[OutputStyle]=="Series",
			
			Prefact=1;
			If[even==2,
				out=Simplify[Sum[A[[l+1,k+1]]HermiteH[k, x/xi]/Sqrt[2^k]/(k!)(coupling)^l,{l,lmin,lmax},{k,0,nu+3lmax}]];
					If[OptionValue[Prefactor]==True, Prefact=Exp[-x^2/xi^2/2+I x^2 alpha /2],Prefact=1],
				out=Sum[A[[l+1,k+1]]x^k coupling^l,{l,lmin,lmax},{k,0,nu+3lmax}];
					If[OptionValue[Prefactor]==True, Prefact=Exp[-x^2 (2 epsilon[[2]])/2],Prefact=1];];
			If[OptionValue[SeriesOutput]=="Coupling",Return[Prefact Collect[out,coupling,Simplify]];];
			If[OptionValue[SeriesOutput]=="Argument",Return[Prefact Collect[out,x,Simplify]];];
		];

(*	If[OptionValue[OutputStyle]=="HighestPower",
		If[OptionValue[ArrayForm]==False,
			out=Sum[A[[l+even l+1,nu+(3+L)l-OptionValue[HighestPower]+1]]x^(nu+(3+L)l-OptionValue[HighestPower])coupling^(l+even l),{l,Ceiling[(-nu+OptionValue[HighestPower])/(3+L)],Floor[lmax-lmax/2 even]-1}];
		];
		If[OptionValue[ArrayForm]==True,
			out=Table[A[[l+even l+1,nu+(3+L)l-OptionValue[HighestPower]+1]],{l,0,Floor[lmax-lmax/2 even]}];
		];
	];*)	
		If[OptionValue[OutputStyle]=="Array",out=Drop[A,lmin]];
		Return[out];
];

];

SetAttributes[LevelPoly,HoldAll];
LevelPoly[Epsilon_,nu_:\[Nu]]:=Module[{a,Rank},
	Rank=Length[Epsilon]-1;
	slvarg=Table[Epsilon[[nu+1]]==Sum[a[k]nu^k,{k,1,Rank}]+a[0],{nu,0,Rank}];
	slv=Solve[slvarg,Table[a[k],{k,0,Rank}]];
	Return[Evaluate[(Sum[a[k]nu^k,{k,1,Rank}]+a[0])/.slv[[1]]]];
];


BWLevelPolynomial[Levels_,nu_:\[Nu]]:=Module[{Out,poly,polycheck},
	LevelsTrans=Transpose[Levels];
	lmax=Length[LevelsTrans]-1;
	poly=Table[LevelPoly[LevelsTrans[[l+1]],nu],{l,0,lmax}];
	polycheck=Table[LevelPoly[Drop[LevelsTrans[[l+1]],-1],nu],{l,0,lmax}];
	check=Simplify[poly-polycheck];
	unreliable=Length[Flatten[Drop[Split[check],1]]];
	Out=Drop[polycheck,-unreliable];
	If[unreliable>0,Print["WARNING: The polynomial form could not be established for the last ",unreliable," orders. \nOutput is truncated to terms which are reliably found." ];];
	Return[Evaluate[Out]];
];



optmain={OutputStyle->"Array",SeriesOutput->"Coupling",Evaluation->"Analytical",Coupling->"g",
HighestPower->0,ArrayForm->False,Echo->False,Monitor->True};
Options[BWDifference]=Join[optmain,{Output->"BWProcess",Evaluation->"Analytical",WorkingPrecision->MachinePrecision,Imaginary->False}];
SetAttributes[BWDifference,HoldAll];
SyntaxInformation[BWDifference]={"ArgumentsPattern"->{_,{_,_},_,_,OptionsPattern[]},"LocalVariables"->{"Solve",{2}}};
BWDifference[DiffOperator_,{P_,X_},nu_,maxhbarord_,OptionsPattern[]]:=Module[ {br,bi,Farrayr, Farrayi, lmax,Farray,MinPowP,MinPowX,MaxPowP,MaxPowX,DiffOPSeries,DiffCoeff,Aconst,Bconst,Cconst,Xiconst,Alphaconst,ScaleFactor,beta,barbeta,MM,Klmax,e,A,Ai,Ar,PowPlist,PowXlist,s,r,karray,Time0,shift,mainoutput,gamma},
(*MinPowP = -1;
MinPowX = -1;
MaxPowP = 1;
MaxPowX = 1;*)
(*P=PX[[1]];
X=PX[[2]];*)
lmax=2*maxhbarord+1;

MinPowP = Normal[Series[P  D[Log[DiffOperator], P], {P, 0, 0}]];
MinPowX = Normal[Series[X  D[Log[DiffOperator], X], {X, 0, 0}]];
MaxPowP = Normal[Series[P  D[Log[DiffOperator], P], {P, Infinity, 0}]];
MaxPowX = Normal[Series[X  D[Log[DiffOperator], X], {X, Infinity, 0}]];
(*Return[{MinPowP,MinPowX,MinPowP,MinPowX}];*)
DiffOPSeries = 
  Series[DiffOperator, {X, 0, MaxPowX}, {P, 0, MaxPowP}];
DiffCoeff = 
  Table[SeriesCoefficient[DiffOPSeries, {X, 0, r}, {P, 0, s}], {r, 
    MinPowX, MaxPowX}, {s, MinPowP, MaxPowP}];
PowPlist = Table[s, {s, MinPowP, MaxPowP}];
PowXlist = Table[r, {r, MinPowX, MaxPowX}];
Aconst = Total[(PowXlist^2).DiffCoeff]/2;
Bconst = Total[DiffCoeff.(PowPlist^2)]/2;
Cconst = (PowXlist).DiffCoeff.(PowPlist);
Xiconst = (4 Bconst^2/(4 Aconst Bconst - Cconst^2))^(1/4);
Alphaconst = -Cconst/(2 Bconst);
ScaleFactor = Sqrt[4 Aconst Bconst - Cconst^2]/2;
beta = Table[(Alphaconst s + r) + I s/Xiconst^2, {s, MinPowP, 
     MaxPowP}, {r, MinPowX, MaxPowX}] 1;
barbeta = 
  Table[(Alphaconst s + r) - I s/Xiconst^2, {s, MinPowP, 
     MaxPowP}, {r, MinPowX, MaxPowX}] 1;
MM = 3;
Klmax = MM lmax + nu + 1;
If[OptionValue[Evaluation]=="Numerical",
	beta*=N[1,OptionValue[WorkingPrecision]];
	barbeta*=N[1,OptionValue[WorkingPrecision]];Xiconst*=N[1,OptionValue[WorkingPrecision]]
	];

If[OptionValue[Imaginary]==False,
Farray[n1_,n2_]:=Farray[n1,n2]=Module[{exp,array,output},
	exp=Tr[(If[n1>0,(beta^n1),ConstantArray[1,Dimensions[beta]]]
			*If[n2>0,barbeta^n2,ConstantArray[1,Dimensions[barbeta]]]).DiffCoeff];
			(*exp=ComplexExpand[Re[exp]];*)
			exp=ComplexExpand[exp];
	If[exp==0,Return[ConstantArray[0,Klmax+1]];];
	array=Table[
				If[k-n1>=0,Hypergeometric2F1[-n1,-n2,1+k-n1,1/2]k!/(k-n1)!/n1!/n2!, 
					2^(k-n1) Hypergeometric2F1[-k,-k+n1-n2,1-k+n1,1/2]/(n1-k)!/(k-n1+n2)!]
				,{k,0,Klmax}];
	exp=Simplify[exp/(2 Bconst/Xiconst^4) ];
	output=exp array;
	Return[output];
];

(*Farray[n1_,n2_]:=Farray[n1,n2]=Module[{exp,array,k},
	exp=Tr[(If[n1>0,
		(beta^n1),ConstantArray[1,Dimensions[beta]]]*If[n2>0,(barbeta)^n2,
		ConstantArray[1,Dimensions[barbeta]]]).DiffCoeff];
		(*exp=Simplify[exp];*)
		Print["I'm here:",exp];
	If[exp!=0,
(*		exp=exp/(2Bconst/Xiconst^4);*)
		array=Table[	If[k-n1>=0,Hypergeometric2F1[-n1,-n2,1+k-n1,1/2]k!/(k-n1)! 1/n1!/n2!, 
					2^(k-n1)Hypergeometric2F1[-k, -k + n1 - n2, 1 - k + n1, 1/2]/(n1-k)!/(k-n1+n2)!],{k,0,Klmax}];
		Print["exp array=",exp array];
		Return[exp array],
		Return[ConstantArray[0,Klmax+1]];
		];
];*)
(**PadLeft[Table[Hypergeometric2F1[-n1,-n2,1+k-n1,1/2]k!/(k-n1)!,{k,0,Klmax}],Klmax+1];*)


Clear[A, e];
e = ConstantArray[0, lmax + 1];
A = ConstantArray[0, {lmax + 1, Klmax + 1}];
A[[1, nu + 1]] = 1;
e[[1]] = 2 nu + 1;
SetSystemOptions["CompileOptions" -> {"SumCompileLength" -> 500}];
karray = Table[If[k != nu, 1/(2 k - 2 nu), 0], {k, 0, Klmax}];
Time0 = AbsoluteTime[]; 
(*Print["over here right:", Farray[0,2]];*)
Monitor[
	Do[
    A[[l + 1]] = 
     karray (-Sum[
     	Farray[n1,n2] If[ (n1 + n2 >= 3),
     			If[ n2 > n1,
     				PadRight[Drop[A[[l + 2 - n1 - n2  + 1]], n2-n1],Klmax + 1],
            		PadLeft[Drop[A[[l + 2 - n1 - n2  + 1]], -n1+n2],Klmax + 1]
            		],
            	0]
            	,{n1,0,l+2}, {n2, 0, l+2-n1}] + 
        Total[Drop[Drop[e, -(lmax - l)] Reverse[Drop[A, -(lmax - l)]],
           1]]);
           A[[l + 1]]=ComplexExpand[Simplify[A[[l+1]]]];
      (*     	Print["A",l,"=",A[[l+1]]];*)
    If[Mod[l, 2] == 0,
      		e[[l + 1]] = Sum[Farray[n1,n2][[nu+1]] If[(n1 + n2 >= 3),
     					A[[l + 2 - n1 - n2  + 1, n2 - n1 + nu + 1]],0]
     					,{n1,0,l+2}, {n2, 0, l+2-n1}];
     		e[[l + 1]]=ComplexExpand[Simplify[e[[l + 1]]]];
     	(*	Print["e",l,"=",e[[l+1]]];*)
     	];
  
    	, {l, 1, lmax}];, 
    "Computing Order: "<>ToString[Floor[l/2]]<>"\nTime lapsed (in seconds): "<>ToString[AbsoluteTime[] - Time0]];
mainoutput={Join[{DiffOperator/.{X->1,P->1}},(e Table[1/Sqrt[2/Xiconst^2]^(i) ScaleFactor, {i, 0, lmax}])[[;; ;;2]]],A Table[1/Sqrt[2/Xiconst^2]^(i), {i, 0, lmax}],{ToExpression["x"], nu,lmax-1,2,Xiconst,Alphaconst}};
(*Return[{(e Table[1/Sqrt[2/Xiconst^2]^(i) ScaleFactor, {i, 0, lmax}])[[;; ;;2]],A,{ToExpression["x"],nu,lmax,0}}];*)
];

If[OptionValue[Imaginary]==True,
gamma=Xiconst^2;
br[n_]:=br[n]=If[n==0,ConstantArray[1,Dimensions[beta]],ComplexExpand[(beta^n+barbeta^n)/2]];
bi[n_]:=bi[n]=If[n==0,ConstantArray[0,Dimensions[beta]],gamma ComplexExpand[(beta^n-barbeta^n)/(2 I)]];

Farrayr[n1_,n2_]:=Farrayr[n1,n2]=Module[{exp,array,output,ng,nl,check},
    If[n1>= n2,ng=n1;nl=n2;,ng=n2;nl=n1;];
	exp=Tr[(If[nl>0,br[nl]^2+bi[nl]^2/gamma^2,ConstantArray[1,Dimensions[beta]]]
				*If[ng-nl>0,br[ng-nl],ConstantArray[1,Dimensions[barbeta]]]).DiffCoeff];
			(*exp=ComplexExpand[Re[exp]];*)
			exp=ComplexExpand[exp];
	If[exp==0,Return[ConstantArray[0,Klmax+1]];];
	array=Table[
				If[k-n1>=0,Hypergeometric2F1[-n1,-n2,1+k-n1,1/2]k!/(k-n1)!/n1!/n2!, 
					2^(k-n1) Hypergeometric2F1[-k,-k+n1-n2,1-k+n1,1/2]/(n1-k)!/(k-n1+n2)!]
				,{k,0,Klmax}];
	exp=Simplify[exp/(2 Bconst/Xiconst^4) ];
	output=exp array;
	Return[output];

];


Farrayi[n1_,n2_]:=Farrayi[n1,n2]=Module[{exp,array,output,ng,nl,sign},
    If[n1>= n2,ng=n1;nl=n2;sign=1,ng=n2;nl=n1;sign=-1;];
	exp=sign Tr[(If[nl>0,br[nl]^2+bi[nl]^2/gamma^2,ConstantArray[1,Dimensions[beta]]]
				*If[ng-nl>0,bi[ng-nl],ConstantArray[0,Dimensions[barbeta]]]).DiffCoeff];
			(*exp=ComplexExpand[Re[exp]];*)
			exp=ComplexExpand[exp];
	If[exp==0,Return[ConstantArray[0,Klmax+1]];];
	array=Table[
				If[k-n1>=0,Hypergeometric2F1[-n1,-n2,1+k-n1,1/2]k!/(k-n1)!/n1!/n2!, 
					2^(k-n1) Hypergeometric2F1[-k,-k+n1-n2,1-k+n1,1/2]/(n1-k)!/(k-n1+n2)!]
				,{k,0,Klmax}];
	exp=Simplify[exp/(2 Bconst/Xiconst^4) ];
	output=exp array;
	Return[output];

];


(*Farray[n1_,n2_]:=Farray[n1,n2]=Module[{exp,array,k},
	exp=Tr[(If[n1>0,
		(beta^n1),ConstantArray[1,Dimensions[beta]]]*If[n2>0,(barbeta)^n2,
		ConstantArray[1,Dimensions[barbeta]]]).DiffCoeff];
		(*exp=Simplify[exp];*)
		Print["I'm here:",exp];
	If[exp!=0,
(*		exp=exp/(2Bconst/Xiconst^4);*)
		array=Table[	If[k-n1>=0,Hypergeometric2F1[-n1,-n2,1+k-n1,1/2]k!/(k-n1)! 1/n1!/n2!, 
					2^(k-n1)Hypergeometric2F1[-k, -k + n1 - n2, 1 - k + n1, 1/2]/(n1-k)!/(k-n1+n2)!],{k,0,Klmax}];
		Print["exp array=",exp array];
		Return[exp array],
		Return[ConstantArray[0,Klmax+1]];
		];
];*)
(**PadLeft[Table[Hypergeometric2F1[-n1,-n2,1+k-n1,1/2]k!/(k-n1)!,{k,0,Klmax}],Klmax+1];*)


Clear[A, e];
e = ConstantArray[0, lmax + 1];
Ar = ConstantArray[0, {lmax + 1, Klmax + 1}];
Ai = ConstantArray[0, {lmax + 1, Klmax + 1}];
Ar[[1, nu + 1]] = 1;
e[[1]] = 2 nu + 1;
SetSystemOptions["CompileOptions" -> {"SumCompileLength" -> 500}];
karray = Table[If[k != nu, 1/(2 k - 2 nu), 0], {k, 0, Klmax}];
Time0 = AbsoluteTime[]; 
(*Print["over here right:", Farray[0,2]];*)
Monitor[
	Do[
    Ar[[l + 1]] = 
     karray (-Sum[
     	Farrayr[n1,n2] If[ (n1 + n2 >= 3),
     			If[ n2 > n1,
     				PadRight[Drop[Ar[[l + 2 - n1 - n2  + 1]], n2-n1],Klmax + 1],
            		PadLeft[Drop[Ar[[l + 2 - n1 - n2  + 1]], -n1+n2],Klmax + 1]
            		],
            	0]-Farrayi[n1,n2] If[ (n1 + n2 >= 3),
     			If[ n2 > n1,
     				PadRight[Drop[Ai[[l + 2 - n1 - n2  + 1]], n2-n1],Klmax + 1],
            		PadLeft[Drop[Ai[[l + 2 - n1 - n2  + 1]], -n1+n2],Klmax + 1]
            		],
            	0]/gamma^2
            	,{n1,0,l+2}, {n2, 0, l+2-n1}] + 
        Total[Drop[Drop[e, -(lmax - l)] Reverse[Drop[Ar, -(lmax - l)]],
           1]]);
           Ar[[l + 1]]=ComplexExpand[Simplify[Ar[[l+1]]]];
 Ai[[l + 1]] = 
     karray (-Sum[
     	Farrayr[n1,n2] If[ (n1 + n2 >= 3),
     			If[ n2 > n1,
     				PadRight[Drop[Ai[[l + 2 - n1 - n2  + 1]], n2-n1],Klmax + 1],
            		PadLeft[Drop[Ai[[l + 2 - n1 - n2  + 1]], -n1+n2],Klmax + 1]
            		],
            	0]+Farrayi[n1,n2] If[ (n1 + n2 >= 3),
     			If[ n2 > n1,
     				PadRight[Drop[Ar[[l + 2 - n1 - n2  + 1]], n2-n1],Klmax + 1],
            		PadLeft[Drop[Ar[[l + 2 - n1 - n2  + 1]], -n1+n2],Klmax + 1]
            		],
            	0]
            	,{n1,0,l+2}, {n2, 0, l+2-n1}] + 
        Total[Drop[Drop[e, -(lmax - l)] Reverse[Drop[Ai, -(lmax - l)]],
           1]]);
           Ai[[l + 1]]=ComplexExpand[Simplify[Ai[[l+1]]]];
      (*     	Print["A",l,"=",A[[l+1]]];*)
    If[Mod[l, 2] == 0,
      		e[[l + 1]] = Sum[Farrayr[n1,n2][[nu+1]] If[(n1 + n2 >= 3),
     					Ar[[l + 2 - n1 - n2  + 1, n2 - n1 + nu + 1]],0]-Farrayi[n1,n2][[nu+1]] If[(n1 + n2 >= 3),
     					Ai[[l + 2 - n1 - n2  + 1, n2 - n1 + nu + 1]],0]/gamma^2
     					,{n1,0,l+2}, {n2, 0, l+2-n1}];
     		e[[l + 1]]=ComplexExpand[Simplify[e[[l + 1]]]];
		
     	(*	Print["e",l,"=",e[[l+1]]];*)
     	];
  
    	, {l, 1, lmax}];, 
    "Computing Order: "<>ToString[Floor[l/2]]<>"\nTime lapsed (in seconds): "<>ToString[AbsoluteTime[] - Time0]];
(*e=Join[{DiffOperator/.{X\[Rule]1,P\[Rule]1}},e];*)
mainoutput={Join[{DiffOperator/.{X->1,P->1}},(e Table[1/Sqrt[2/Xiconst^2]^(i) ScaleFactor, {i, 0, lmax}])[[;; ;;2]]],(Ar+I Ai)Table[1/Sqrt[2/Xiconst^2]^(i), {i, 0, lmax}],{ToExpression["x"], nu,lmax-1,2,Xiconst,Alphaconst}};

];

Return[BWProcess[mainoutput,Output->OptionValue[Output],OutputStyle->OptionValue[OutputStyle],SeriesOutput->OptionValue[SeriesOutput],Evaluation->OptionValue[Evaluation],Coupling->OptionValue[Coupling],HighestPower->OptionValue[HighestPower],ArrayForm->OptionValue[ArrayForm]]];
];


(* Wolfram Language Raw Program *)




End[];
EndPackage[ ]












